#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef LINUX
#include "conion.h"
#else
#include <conio.h>
#endif
#include <ctype.h>
#include "interp.h"

#define OP_NONE                      0
#define OP_UNSIGNED_OBJ_OFFSET       1
#define OP_UNSIGNED_VAR_OFFSET       2
#define OP_UNSIGNED_LOC_OFFSET       3
#define OP_DATA                      4
#define OP_SIGNED_JMP_OFFSET         5
#define OP_MEMORY_OPCODE_WRITE       6
#define OP_OBJ_CALL_PAIR             7
#define OP_SIGNED_OFFSET             8
#define OP_PACKED_LITERAL            9
#define OP_BYTE_LITERAL             10
#define OP_WORD_LITERAL             11
#define OP_NEAR_LONG_LITERAL        12
#define OP_LONG_LITERAL             13
#define OP_MEMORY_OPCODE            14
#define OP_MEMORY_OPCODE_READ       15
#define OP_COMPACT_VAR_OFFSET       16
#define OP_COMPACT_LOC_OFFSET       17

#define OS_COMMAND 0x7ff8
#define OS_PARM    0x7ffc

extern char hubram[0x10000];
static char ops[3000];

extern int symflag;

void RemoveCRLF(char *str)
{
    int len = strlen(str);

    if(len == 0) return;

    str += len-1;

    while (len > 0)
    {
	if (*str != 10 && *str != 13) break;
	*str-- = 0;
	len--;
    }
}

void ReadOpsFile(void)
{
    FILE *infile;
    char *ptr = ops;

    infile = fopen("opsfile.txt", "r");

    if (!infile)
    {
	printf("Could not open opsfile.txt\n");
	exit(1);
    }

    while (fgets(ptr, 100, infile))
    {
	RemoveCRLF(ptr);
	//printf("ops %d: %s\n", i, ptr);
	ptr += strlen(ptr) + 1;
    }
    *ptr = 0;
    fclose(infile);
}

char *FindOpcode(int pcurr, int *ops_format, int mode)
{
    char *ops_str = ops;
    int opcode, opcode1;

    opcode = BYTE(pcurr);

    if (mode)
    {
	opcode &= 0x7f;
	if (opcode >= 0x40)
	    opcode += 0xe0 - 0x40;
	else if (opcode >= 0x20)
	    opcode &= 0x78;
	else if (opcode >= 0x08)
	    opcode &= 0x7c;
	else
	    opcode &= 0x7e;
    }
    else
    {
        if (opcode >= 0x40 && opcode <= 0x7f) opcode &= 0x63;
    }

    while (*ops_str)
    {
        sscanf(ops_str+3, "%x", &opcode1);
	//printf("FindOpcode: %2.2x %s\n", opcode1, ops_str);
	if ((opcode == opcode1) && ((mode == 0) || (ops_str[0] >= '2'))) break;
	ops_str += strlen(ops_str) + 1;
    }

    if(!(*ops_str)) return 0;
#if 0
    if (opcode == 0x3f)
    {
	if (BYTE(pcurr+1) & 0x20)
	    ops_str += strlen(ops_str) + 1;
	if (BYTE(pcurr+1) & 0x40)
	    ops_str += strlen(ops_str) + 1;
    }
#endif
    if (ops_str[1] == ' ') *ops_format = 0;
    else *ops_format = ops_str[1] - '@';
    if (ops_str[0] != ' ') *ops_format |= (ops_str[0] - '0') << 5;
    return ops_str + 6;
}

extern char objname[100][20];
extern int methodnum[100];
extern int methodlev;

char *FindChar(char *str, int val)
{
    while (*str && *str != val)
        str++;
    return str;
}

void ProcessRet(void)
{
    if (!symflag) return;
    methodlev--;
    printf("return %s\n\n", objname[methodlev]);
}

static char linebuf[200];

void ProcessCall(int subnum, int mode)
{
    int methnum = 0;
    FILE *infile;

    if (!symflag) return;

    infile = fopen(objname[methodlev], "r");
    //printf("ProcessCall: %d, %d, %s\n", subnum, methodlev, objname[methodlev]);
    if (mode)
    {
	methodlev++;
        strcpy(objname[methodlev], objname[methodlev-1]);
    }
    methodnum[methodlev] = subnum;
    if (!infile) return;
    // Count pubs
    while (fgets(linebuf, 200, infile))
    {
	if (strncmp(linebuf, "PUB", 3) == 0 || strncmp(linebuf, "pub", 3) == 0)
	{
	    methnum++;
	    if (methnum == subnum)
	    {
		printf("call %s:%s\n", objname[methodlev], linebuf);
		fclose(infile);
		return;
	    }
	}
    }
    fclose(infile);
    infile = fopen(objname[methodlev], "r");
    // Count pris
    while (fgets(linebuf, 200, infile))
    {
	if (strncmp(linebuf, "PRI", 3) == 0 || strncmp(linebuf, "pri", 3) == 0)
	{
	    methnum++;
	    if (methnum == subnum)
	    {
		printf("call %s:%s\n", objname[methodlev], linebuf);
		fclose(infile);
		return;
	    }
	}
    }
}

void ProcessObjCall(int objnum, int subnum)
{
    int mode = 0;
    int methnum = 0;
    int commentmode = 0;
    int found = 0;
    FILE *infile;

    if (!symflag) return;

    infile = fopen(objname[methodlev], "r");
    //printf("ProcessObjCall: %d, %d, %d, %s\n", objnum, subnum, methodlev, objname[methodlev]);
    methodlev++;
    strcpy(objname[methodlev], objname[methodlev-1]);
    methodnum[methodlev] = subnum;
    if (!infile) return;
    // Count pubs and pris
    while (fgets(linebuf, 200, infile))
    {
	if (linebuf[0] == '}')
	{
	    commentmode--;
	    if (commentmode < 0) commentmode = 0;
	    continue;
	}
	else if (linebuf[0] == '{')
	{
	    if (*FindChar(linebuf, '}') == 0)
	        commentmode++;
	}
	if (commentmode) continue;
	if (strncmp(linebuf, "PUB", 3) == 0 || strncmp(linebuf, "pub", 3) == 0 ||
	    strncmp(linebuf, "PRI", 3) == 0 || strncmp(linebuf, "pri", 3) == 0)
	{
	    methnum++;
	}
    }
    fclose(infile);
    infile = fopen(objname[methodlev], "r");
    // Count objs
    commentmode = 0;
    while (fgets(linebuf, 200, infile))
    {
	if (linebuf[0] == '}')
	{
	    commentmode--;
	    if (commentmode < 0) commentmode = 0;
	    continue;
	}
	else if (linebuf[0] == '{')
	{
	    if (*FindChar(linebuf, '}') == 0)
	        commentmode++;
	}
	if (commentmode) continue;
	if (mode == 0)
	{
	    if (strncmp(linebuf, "OBJ", 3) == 0 || strncmp(linebuf, "obj", 3) == 0)
	        mode = 1;
	}
	else
	{
	    methnum++;
	    //printf("%d %s", methnum, linebuf);
	    if (linebuf[0] != ' ') break;
	    if (methnum == objnum)
	    {
		char *ptr1;
		char *ptr2;
		int num;

		// Locate filename
		ptr1 = FindChar(linebuf, '"');
		if (*ptr1)
		    ptr1++;
		    ptr2 = FindChar(ptr1, '"');
		    if (*ptr2) ptr2--;
		else
		    ptr2 = ptr1;
		num = ptr2 - ptr1 + 1;
		if (num < 1) num = 1;
		memcpy(objname[methodlev], ptr1, num);
		objname[methodlev][num] = 0;
		strcat(objname[methodlev], ".spn");
		//printf("%s\n", objname[methodlev]);
		fclose(infile);
		found = 1;
		break;
	    }
	}
    }
    if (!found)
    {
        fclose(infile);
	return;
    }
    ProcessCall(subnum, 0);
}

void PrintOp(CogVarsT *cogvars)
{
    long pcurr = cogvars->pcurr;
    int opcode, opform;
    char *opstr;
    int val;
    int operand;
    char *regop[] = {"ldreg", "streg", "exreg", "??reg"};
    char *regname[] = {"par", "cnt", "ina", "inb", "outa", "outb", "dira",
	"dirb", "ctra", "ctrb", "frqa", "frqb", "phsa", "phsb", "vcfg", "vscl"};

    if (cogvars->state != 1) return;

    //printf("Cog %d: ", cogvars->id);

    opcode = BYTE(pcurr);
    //printf("PrintOp: %4.4x %2.2x\n", pcurr, opcode);
    opstr = FindOpcode(pcurr, &opform, 0);
    switch (opform & 0x1f)
    {
        case OP_NONE:
	    printf("%4.4x: %2.2x             %s\n", pcurr, opcode, opstr);
	    if (strncmp(opstr, "ret", 3) == 0)
	        ProcessRet();
	    //if (strcmp(opstr, "ret") == 0)
	        //printf("\n");
            pcurr++;
            break;
        case OP_UNSIGNED_OBJ_OFFSET:
        case OP_UNSIGNED_VAR_OFFSET:
        case OP_UNSIGNED_LOC_OFFSET:
	    operand = BYTE(pcurr+1);
	    if (operand & 0x80)
	    {
	        operand = ((operand & 0x7f) << 8) | BYTE(pcurr+2);
		printf("%4.4x: %2.2x %2.2x %2.2x       %s $%x\n", pcurr, opcode, BYTE(pcurr+1), BYTE(pcurr+2),  opstr, operand);
                pcurr += 3;
	    }
	    else
	    {
		printf("%4.4x: %2.2x %2.2x          %s $%x\n", pcurr, opcode, BYTE(pcurr+1),  opstr, operand);
                pcurr += 2;
	    }
            break;
        case OP_DATA:
	    printf("%4.4x: %2.2x %s - ******* TBD *******\n", pcurr, opcode, opstr);
            pcurr++;
            break;
        case OP_SIGNED_OFFSET:
        case OP_SIGNED_JMP_OFFSET:
	    operand = BYTE(pcurr+1);
	    if (operand & 0x80)
	    {
	        operand = ((operand & 0x7f) << 8) | BYTE(pcurr+2);
		operand = (operand << 17) >> 17;
		printf("%4.4x: %2.2x %2.2x %2.2x       %s %d\n", pcurr, opcode, BYTE(pcurr+1), BYTE(pcurr+2),  opstr, operand);
                pcurr += 3;
	    }
	    else
	    {
		operand = (operand << 25) >> 25;
		printf("%4.4x: %2.2x %2.2x          %s %d\n", pcurr, opcode, BYTE(pcurr+1),  opstr, operand);
                pcurr += 2;
	    }
            break;
        case OP_MEMORY_OPCODE_WRITE:
	    operand = (BYTE(pcurr+1) & 31);
	    if (operand <= 15)
	    {
	        printf("%4.4x: %2.2x %2.2x          %s $%3.3x\n", pcurr, opcode, BYTE(pcurr+1), opstr, operand + 0x1e0);
	    }
	    else
	    {
		printf("%4.4x: %2.2x %2.2x          %s %s\n", pcurr, opcode, BYTE(pcurr+1), opstr, regname[operand-16]);
	    }
            pcurr += 2;
            break;
        case OP_OBJ_CALL_PAIR:
	    printf("%4.4x: %2.2x %2.2x %2.2x       %s %d %d\n", pcurr, opcode, BYTE(pcurr+1), BYTE(pcurr+2), opstr, BYTE(pcurr+1), BYTE(pcurr+2));
	    ProcessObjCall(BYTE(pcurr+1), BYTE(pcurr+2));
            pcurr += 3;
            break;
        case OP_PACKED_LITERAL:
	    operand = BYTE(pcurr+1);
	    val = 2 << (operand & 31);
	    if (operand & 0x20) val--;
	    if (operand & 0x40) val = ~val;
	    printf("%4.4x: %2.2x %2.2x          %s $%x\n", pcurr, opcode, operand, opstr, val);
            pcurr += 2;
            break;
        case OP_BYTE_LITERAL:
	    operand = BYTE(pcurr+1);
	    printf("%4.4x: %2.2x %2.2x          %s %d\n", pcurr, opcode, operand, opstr, operand);
	    if (strcmp(opstr, "call") == 0)
	        ProcessCall(operand, 1);
	        //printf("Process call\n");
	    //if (strcmp(opstr, "call") == 0)
	        //printf("\n");
            pcurr += 2;
            break;
        case OP_WORD_LITERAL:
	    operand = BYTE(pcurr+1);
	    operand = (operand << 8) | BYTE(pcurr+2);
	    printf("%4.4x: %2.2x %2.2x %2.2x       %s %d\n", pcurr, opcode, BYTE(pcurr+1), BYTE(pcurr+2),  opstr, operand);
            pcurr += 3;
            break;
        case OP_NEAR_LONG_LITERAL:
	    operand = BYTE(pcurr+1);
	    operand = (operand << 8) | BYTE(pcurr+2);
	    operand = (operand << 8) | BYTE(pcurr+3);
	    printf("%4.4x: %2.2x %2.2x %2.2x %2.2x    %s %d\n", pcurr, opcode,
	        BYTE(pcurr+1), BYTE(pcurr+2), BYTE(pcurr+3), opstr, operand);
            pcurr += 4;
            break;
        case OP_LONG_LITERAL:
	    operand = BYTE(pcurr+1);
	    operand = (operand << 8) | BYTE(pcurr+2);
	    operand = (operand << 8) | BYTE(pcurr+3);
	    operand = (operand << 8) | BYTE(pcurr+4);
	    printf("%4.4x: %2.2x %2.2x %2.2x %2.2x %2.2x %s %d\n", pcurr, opcode,
	        BYTE(pcurr+1), BYTE(pcurr+2), BYTE(pcurr+3), BYTE(pcurr+4), opstr, operand);
            pcurr += 5;
            break;
        case OP_MEMORY_OPCODE:
	    printf("%4.4x: %2.2x %s - ******* TBD *******\n", pcurr, opcode, opstr);
            pcurr++;
            break;
        case OP_MEMORY_OPCODE_READ:
	    operand = (BYTE(pcurr+1) >> 5) & 3;
	    opstr = regop[operand];
	    if (operand == 2) opform |= 1 << 5;
	    operand = (BYTE(pcurr+1) & 31);
	    if (operand <= 15)
	    {
                printf("%4.4x: %2.2x %2.2x          %s $%3.3x\n", pcurr, opcode, BYTE(pcurr+1), opstr, operand + 0x1e0);
	    }
	    else
	    {
                printf("%4.4x: %2.2x %2.2x          %s %s\n", pcurr, opcode, BYTE(pcurr+1), opstr, regname[operand-16]);
	    }
            pcurr += 2;
            break;
        case OP_COMPACT_VAR_OFFSET:
	    operand = opcode & 0x1c;
	    printf("%4.4x: %2.2x             %s $%x\n", pcurr, opcode, opstr, operand);
            pcurr++;
            break;
        case OP_COMPACT_LOC_OFFSET:
	    operand = opcode & 0x1c;
	    printf("%4.4x: %2.2x             %s $%x\n", pcurr, opcode, opstr, operand);
            pcurr++;
            break;
    }
    if ((opform >> 5) == 1)
    {
	char *loadstr = "";
	opcode = BYTE(pcurr);
	if (opcode & 0x80)
	{
	    loadstr = "load";
	    opcode &= 0x7f;
	}
        opstr = FindOpcode(pcurr, &opform, 1);
        //printf("Cog %d: ", cogvars->id);
	if (opcode != 0x02)
	    printf("%4.4x: %2.2x             %s %s\n", pcurr, BYTE(pcurr), opstr, loadstr);
	else
	{
	    operand = BYTE(pcurr+1);
	    if (operand < 0x80)
	    {
	        operand = (operand << 25) >> 25;
	        printf("%4.4x: %2.2x %2.2x          %s %s %d\n", pcurr, BYTE(pcurr), BYTE(pcurr+1), opstr, loadstr, operand);
	        pcurr++;
	    }
	    else
	    {
	        operand = (operand << 8) | BYTE(pcurr+2);
	        operand = (operand << 17) >> 17;
	        printf("%4.4x: %2.2x %2.2x %2.2x       %s %s %d\n", pcurr, BYTE(pcurr), BYTE(pcurr+1), BYTE(pcurr+2), opstr, loadstr, operand);
	        pcurr += 2;
	    }
	}
	pcurr++;
    }
}
