{{
  This object is used to test SpinSim.  It performs console I/O, and opens and reads files.
  Added a "run" command to load a binary Spin file and execute it.
  Added a "dir" command to list the directory
}}
obj
  ser : "conio"
  fs : "fileio"

var
  byte buffer[100]
  
pub main | tokens[10], numtokens
  fs.mount(16)
  ser.start(31, 30, 0, 115200)
  
  help
  
  repeat
    ser.tx(13)
    ser.tx(">")
    gets(@buffer)
    numtokens := tokenize(@buffer, @tokens)
    ifnot numtokens
      next
    if strcomp(tokens[0], string("exit"))
      quit
    elseif strcomp(tokens[0], string("help"))
      help
    elseif strcomp(tokens[0], string("dump"))
      if numtokens == 2
        dumpfile(tokens[1])
      else
        help
    elseif strcomp(tokens[0], string("type"))
      if numtokens == 2
        typefile(tokens[1])
      else
        help
    elseif strcomp(tokens[0], string("run"))
      if numtokens == 2
        runfile(tokens[1], 0)
      else
        help
    elseif strcomp(tokens[0], string("run0"))
      if numtokens == 2
        runfile(tokens[1], 1)
      else
        help
    elseif strcomp(tokens[0], string("dir"))
      printdir
    else
      ser.str(string("Unknown command", 13))
      help

pub help
  ser.str(string(13, "Avaliable Commands", 13))
  ser.str(string("help      - Print the available commands", 13))
  ser.str(string("dump file - Dump the contents of a file in hex", 13))
  ser.str(string("type file - Print the contents of a file", 13))
  ser.str(string("run file  - Load and run Spin binary file in a new cog", 13))
  ser.str(string("run0 file - Load and run Spin binary file in the current cog", 13))
  ser.str(string("dir       - Directory listing", 13))
  ser.str(string("exit      - Exit the program", 13))
  
pub typefile(fname) | num, i, value, prev
  value~
  prev~ 
  ifnot fs.popen(fname, "R")
    repeat
      num := fs.pread(@value, 1)
      if num =< 0
        return
      if value == 10
        if prev <> 13
          ser.tx(13)
      else
        ser.tx(value)
      prev := value
  else
    ser.str(string("Could not open "))
    ser.str(fname)
    ser.tx(13)

pub dumpfile(fname) | num, i, addr
  addr~  
  ifnot fs.popen(fname, "R")
    repeat
      num := fs.pread(@buffer, 16)
      if num =< 0
        return
      ser.hex(addr, 4)
      ser.tx(" ")
      addr += 16
      repeat i from 0 to num - 1
        ser.hex(byte[@buffer][i], 2)
        ser.tx(" ")
      ser.tx(13)
  else
    ser.str(string("Could not open "))
    ser.str(fname)
    ser.tx(13)

pub runfile(fname, mode) | num, size, addr, i, dbase
  if fs.popen(fname, "R")
    ser.str(string("Could not open "))
    ser.str(fname)
    ser.tx(13)
    return
  if mode
    fs.pclose
    addr := $7080
    bytemove($7000, fname, strsize(fname) + 1)
    fname := string("loader.binary")
    if fs.popen(fname, "R")
      ser.str(string("Could not open "))
      ser.str(fname)
      ser.tx(13)
      return
  else
    addr := @addr + 800 ' Set starting address 200 longs after stack
  size := $8000 - addr
  num := fs.pread(addr, size)
  if num == size
    ser.str(string("File too large", 13))
    return
  repeat i from 6 to 14 step 2
    word[addr + i] += addr
  dbase := word[addr + 10]
  long[dbase]~
  long[dbase - 4] := $fff9ffff
  long[dbase - 8] := $fff9ffff
  if mode
    coginit(cogid, $f004, addr + 4)
  else
    cognew($f004, addr + 4)

pub printdir | pdir, dirent, num
  pdir := fs.hopendir
  repeat while dirent := fs.hreaddir(pdir)
    if long[dirent][1] & 16
      ser.str(string("<DIR>   "))
    else
      num := long[dirent]
      if num < 1
        num := 1
      repeat while num < 1000000
        ser.tx(" ")
        num *= 10
    ser.dec(long[dirent])
    ser.tx(" ")
    ser.str(@long[dirent][2])
    ser.tx(13)
  fs.hclosedir(pdir)

pub gets(str) | char, str0
  str0 := str
  repeat
    char := ser.rx
    if char == 8
      if str <> str0
        str--
        ser.tx(8)
        ser.tx(" ")
        ser.tx(8)
      next
    ser.tx(char)
    if char == 13 or char == 10
      byte[str++]~
      return    
    byte[str++] := char

pub tokenize(str, tokens)
  repeat 10
    str := skipblanks(str)
    ifnot byte[str]
      return
    long[tokens][result++] := str
    str := findblanks(str)
    if byte[str] == 0
      return
    byte[str++]~

pub findblanks(str)
  repeat while byte[str] and byte[str] <> " "
    str++
  return str
    
pub skipblanks(str)
  repeat while byte[str] and byte[str] == " "
    str++
  return str
    
